/*
 * pty.c
 *
 *   Tcl command and supporting functions to open a pty/tty pair 
 *   and exec a program attached to the pair.
 *
 * Copyright (C) 1996, 1997
 *	Dr. Duncan McLennan Barclay.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY DUNCAN BARCLAY AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL DUNCAN BARCLAY OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *   $Id: pty.c,v 1.6 1997/02/21 22:57:20 dmlb Exp $
 */

#include "tcl.h"
#include "pty.h"

#include <stdlib.h>
#include <string.h>
#define TTYDEFCHARS
#include <termios.h>
#include <unistd.h>
#include <sys/errno.h>
#include <sys/fcntl.h>
#include <sys/ioctl.h>

/*
 * Forward declartions
 */
static int      ptyBlockingMode _ANSI_ARGS_((
                  ClientData instanceData, int mode));
static int      ptyClose _ANSI_ARGS_((
                  ClientData instanceData, Tcl_Interp *interp));
static int      ptyInput _ANSI_ARGS_((
                  ClientData instanceData, char *buf, 
                  int toRead, int *errorCode));
static int      ptyOutput _ANSI_ARGS_((
                  ClientData instanceDate, char *buf,
                  int toWrite, int *errorCode));
static void     ptyWatchChannel _ANSI_ARGS_((
                  ClientData instanceDate, int mask));
static int      ptyChannelReady _ANSI_ARGS_((
                  ClientData instanceDate, int mask));
static Tcl_File ptyGetFile _ANSI_ARGS_((
                  ClientData instanceDate, int direction));

/*
 * Pty structures
 */
static Tcl_ChannelType ptyChannelType = {
  "pty",                /* Type name. */
  ptyBlockingMode,      /* Set blocking/nonblocking mode on master. */
  ptyClose,             /* Close proc. */
  ptyInput,             /* Input proc. */
  ptyOutput,            /* Output proc. */
  NULL,                 /* Seek proc. */
  NULL,                 /* Set option proc. */
  NULL,                 /* Get option proc. */
  ptyWatchChannel,      /* Watch proc. */
  ptyChannelReady,      /* Ready proc. */
  ptyGetFile            /* Get file proc. */
};

typedef struct ptyState {
  char tline[16];
  char pline[16];
  Tcl_File file;
} ptyState;

/*
 * Structures for parsing termios options etc.
 */
struct termios_values {
  char sw[16];
  int value;
};

static struct termios_values imode_elements[] = {  
  {"ignbrk",    IGNBRK},
  {"brkint",    BRKINT},
  {"ignpar",    IGNPAR},
  {"parmrk",    PARMRK},
  {"inpck",     INPCK},
  {"istrip",    ISTRIP},
  {"inlcr",     INLCR},
  {"igncr",     IGNCR},
  {"icrnl",     ICRNL},
  {"ixon",      IXON},
  {"ixoff",     IXOFF},
  {"ixany",     IXANY},
  {"imaxbel",   IMAXBEL}
};
#define IMODE_ELEMENTS (sizeof (imode_elements)/sizeof (struct termios_values))

static struct termios_values omode_elements[] = {  
  {"opost",     OPOST},
  {"onlcr",     ONLCR},
  {"oxtabs",    OXTABS},
  {"onoeot",    ONOEOT}
};
#define OMODE_ELEMENTS (sizeof (omode_elements)/sizeof (struct termios_values))

static struct termios_values cmode_elements[] = {  
  {"cstopb",    CSTOPB},
  {"cread",     CREAD},
  {"parenb",    PARENB},
  {"parodd",    PARODD},
  {"hupcl",     HUPCL},
  {"clocal",    CLOCAL},
  {"ccts_oflow",CCTS_OFLOW},
  {"crtscts",   CRTSCTS},
  {"crts_iflow",CRTS_IFLOW},
  {"mdmbuf",    MDMBUF}
};
#define CMODE_ELEMENTS (sizeof (cmode_elements)/sizeof (struct termios_values))

static struct termios_values lmode_elements[] = {
  {"echoke",    ECHOKE},
  {"echoe",     ECHOE},
  {"echok",     ECHOK},
  {"echo",      ECHO},
  {"echonl",    ECHONL},
  {"echoprt",   ECHOPRT},
  {"echoctl",   ECHOCTL},
  {"isig",      ISIG},
  {"icanon",    ICANON},
  {"altwerase", ALTWERASE},
  {"iexten",    IEXTEN},
  {"extproc",   EXTPROC},
  {"tostop",    TOSTOP},
  {"flusho",    FLUSHO},
  {"nokerninfo",NOKERNINFO},
  {"pendin",    PENDIN},
  {"noflsh",    NOFLSH}
};
#define LMODE_ELEMENTS (sizeof (lmode_elements)/sizeof (struct termios_values))

static struct termios_values cchar_elements[] = {
  {"eof",       VEOF},
  {"eol",       VEOL},
  {"eol2",      VEOL2},
  {"erase",     VERASE},
  {"werase",    VWERASE},
  {"kill",      VKILL},
  {"reprint",   VREPRINT},
  {"intr",      VINTR},
  {"quit",      VQUIT},
  {"susp",      VSUSP},
  {"dsusp",     VDSUSP},
  {"start",     VSTART},
  {"stop",      VSTOP},
  {"lnext",     VLNEXT},
  {"discard",   VDISCARD},
  {"status",    VSTATUS}
};
#define CCHAR_ELEMENTS (sizeof (cchar_elements)/sizeof (struct termios_values))

static struct termios_values baud_elements[] = {
  {"0",        B0},
  {"50",       B50},
  {"75",       B75},
  {"110",      B110},
  {"134",      B134},
  {"150",      B150},
  {"200",      B200},
  {"300",      B300},
  {"600",      B600},
  {"1200",     B1200},
  {"1800",     B1800},
  {"2400",     B2400},
  {"4800",     B4800},
  {"7200",     B7200},
  {"9600",     B9600},
  {"14400",    B14400},
  {"19200",    B19200},
  {"28800",    B28800},
  {"38400",    B38400},
  {"57600",    B57600},
  {"76800",    B76800},
  {"115200",   B115200},
  {"230400",   B230400},
};
#define BAUD_ELEMENTS (sizeof (baud_elements)/sizeof (struct termios_values))

/*
 *----------------------------------------------------------------------
 *
 * ptyBlockingMode--
 *
 *      Helper procedure to set blocking and nonblocking modes on a
 *      channel. Invoked either by generic IO level code or by other
 *      channel drivers after doing channel-type-specific inialization.
 *
 * Results:
 *      0 if successful, errno when failed.
 *
 * Side effects:
 *      Sets the device into blocking or non-blocking mode.
 *
 *----------------------------------------------------------------------
 */
static int
ptyBlockingMode (instanceData, mode)
  ClientData instanceData;            /* Instance of the pty. */
  int mode;                           /* The mode to set. Can be one of
                                       * TCL_MODE_BLOCKING or
                                       * TCL_MODE_NONBLOCKING. */
{
  int curStatus;
  int fd;
  ptyState *ptyPtr;

  ptyPtr = (ptyState *)instanceData;

  fd = (int)Tcl_GetFileInfo (ptyPtr->file, NULL);
  curStatus = fcntl (fd, F_GETFL);
  if (mode == TCL_MODE_BLOCKING) {
    curStatus &= (~(O_NONBLOCK));
  } else {
    curStatus |= O_NONBLOCK;
  }
  if (fcntl (fd, F_SETFL, curStatus) < 0) {
    return errno;
  }

  return 0;
}

/*
 *----------------------------------------------------------------------
 *
 * ptyCloseProc--
 *
 *      This procedure is called from the generic IO level to perform
 *      channel-type-specific cleanup when a pty based channel is closed.
 *
 * Results:
 *      0 if successful, errno if failed.
 *
 * Side effects:
 *      Closes the device of the channel.
 *
 *----------------------------------------------------------------------
 */
static int
ptyClose (instanceData, interp)
  ClientData instanceData;    /* Client data containg device names. */
  Tcl_Interp *interp;         /* For error reporting - unused. */
{
  int fd, errorCode = 0;
  ptyState *ptyPtr;

  ptyPtr = (ptyState *)instanceData;

  fd = (int)Tcl_GetFileInfo (ptyPtr->file, NULL);
  Tcl_FreeFile (ptyPtr->file);

  if (close(fd) < 0) {
    errorCode = errno;
    if (interp != (Tcl_Interp *)NULL) {
      Tcl_AppendResult (interp, "couldn't close pty",
                        (char *)NULL);
    }
  }

  Tcl_ReapDetachedProcs();

  if ((ptyGetPrivs () < 0) && (errorCode == 0)) {
    errorCode = errno;
    if (interp != (Tcl_Interp *)NULL) {
      Tcl_AppendResult (interp, "couldn't restore set uid/gid privilages",
                        (char *)NULL);
    }
  }
  if ((chmod (ptyPtr->tline, 0666) < 0) && (errorCode == 0)) {
    errorCode = errno;
    if (interp != (Tcl_Interp *)NULL) {
      Tcl_AppendResult (interp, "couldn't restore pty permissions",
                        (char *)NULL);
    }
  }
  if ((chown (ptyPtr->tline, 0, 0) < 0) && (errorCode == 0)) {
    errorCode = errno;
    if (interp != (Tcl_Interp *)NULL) {
      Tcl_AppendResult (interp, "couldn't restore pty ownership",
                        (char *)NULL);
    }
  }
  if ((chmod (ptyPtr->pline, 0666) < 0) && (errorCode == 0)) {
    errorCode = errno;
    if (interp != (Tcl_Interp *)NULL) {
      Tcl_AppendResult (interp, "couldn't restore tty permissions",
                        (char *)NULL);
    }
  }
  if ((chown (ptyPtr->pline, 0, 0) < 0) && (errorCode == 0)) {
    errorCode = errno;
    if (interp != (Tcl_Interp *)NULL) {
      Tcl_AppendResult (interp, "couldn't restore tty ownership",
                        (char *)NULL);
    }
  }
  if ((ptyReleasePrivs () < 0) && (errorCode == 0)) {
    errorCode = errno;
    if (interp != (Tcl_Interp *)NULL) {
      Tcl_AppendResult (interp, "couldn't release set uid/gid privilages",
                        (char *)NULL);
    }
  }
  ckfree ((char *)ptyPtr);

  return errorCode;
}

/*
 *----------------------------------------------------------------------
 *
 * ptyInput--
 *
 *      This procedure is invoked from the generic IO level to read
 *      input from the master side of the pty/tty pair.
 *
 * Results:
 *      The number of bytes read is returned or -1 on error. An output
 *      argument contains a POSIX error code if an error occurs, or zero.
 *
 * Side effects:
 *      Reads input from the input device of the channel.
 *
 *----------------------------------------------------------------------
 */
static int
ptyInput (instanceData, buf, toRead, errorCodePtr)
  ClientData instanceData;            /* Instance of this pty. */
  char *buf;                          /* Where to store data read. */
  int toRead;                         /* How much space is available
                                       * in the buffer? */
  int *errorCodePtr;                  /* Where to store error code. */
{
  int fd;                             /* The OS handle for reading. */
  int bytesRead;                      /* How many bytes were actually
                                       * read from the input device? */
  ptyState *ptyPtr;

  *errorCodePtr = 0;
  ptyPtr = (ptyState *)instanceData;
  fd = (int)Tcl_GetFileInfo (ptyPtr->file, NULL);

  /*
   * Assume there is always enough input available. This will block
   * appropriately, and read will unblock as soon as a short read is
   * possible, if the channel is in blocking mode. If the channel is
   * nonblocking, the read will never block.
   */

  bytesRead = read (fd, buf, (size_t) toRead);
  if (bytesRead > -1) {
      return bytesRead;
  }
  *errorCodePtr = errno;
  return -1;
}

/*
 *----------------------------------------------------------------------
 *
 * ptyOutput--
 *
 *      This procedure is invoked from the generic IO level to write
 *      output to the master side of the pty/tty pair.
 *
 * Results:
 *      The number of bytes written is returned or -1 on error. An
 *      output argument contains a POSIX error code if an error occurred,
 *      or zero.
 *
 * Side effects:
 *      Writes output on the output device of the channel.
 *
 *----------------------------------------------------------------------
 */
static int
ptyOutput (instanceData, buf, toWrite, errorCodePtr)
  ClientData instanceData;            /* Instance data for this pty. */
  char *buf;                          /* The data buffer. */
  int toWrite;                        /* How many bytes to write? */
  int *errorCodePtr;                  /* Where to store error code. */
{
  int written;
  int fd;
  ptyState *ptyPtr;

  *errorCodePtr = 0;
  ptyPtr = (ptyState *)instanceData;
  fd = (int)Tcl_GetFileInfo (ptyPtr->file, NULL);

  written = write (fd, buf, (size_t)toWrite);
  if (written > -1) {
      return written;
  }
  *errorCodePtr = errno;
  return -1;
}

/*
 *----------------------------------------------------------------------
 *
 * ptyWatchChannel--
 *
 *	Initialize the notifier to watch Tcl_Files from this channel.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Sets up the notifier so that a future event on the channel will
 *	be seen by Tcl.
 *
 *----------------------------------------------------------------------
 */
static void
ptyWatchChannel (instanceData, mask)
  ClientData instanceData;            /* The file state. */
  int mask;                           /* Events of interest; an OR-ed
                                       * combination of TCL_READABLE,
                                       * TCL_WRITABLE and TCL_EXCEPTION. */
{
  ptyState *ptyPtr;

  ptyPtr = (ptyState *)instanceData;

  Tcl_WatchFile (ptyPtr->file, TCL_READABLE);
  Tcl_WatchFile (ptyPtr->file, TCL_WRITABLE);
  Tcl_WatchFile (ptyPtr->file, TCL_EXCEPTION);
}

/*
 *----------------------------------------------------------------------
 *
 * ptyChannelReady--
 *
 *	Called by the notifier to check whether events of interest are
 *	present on the channel.
 *
 * Results:
 *	Returns OR-ed combination of TCL_READABLE, TCL_WRITABLE and
 *	TCL_EXCEPTION to indicate which events of interest are present.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */
static int
ptyChannelReady (instanceData, mask)
  ClientData instanceData;            /* The file state. */
  int mask;                           /* Events of interest; an OR-ed
                                       * combination of TCL_READABLE,
                                       * TCL_WRITABLE and TCL_EXCEPTION. */
{
  int present = 0;
  ptyState *ptyPtr;

  ptyPtr = (ptyState *)instanceData;

  if (mask & TCL_READABLE) {
    present |= Tcl_FileReady (ptyPtr->file, TCL_READABLE);
  }
  if (mask & TCL_WRITABLE) {
    present |= Tcl_FileReady (ptyPtr->file, TCL_WRITABLE);
  }
  if (mask & TCL_EXCEPTION) {
    present |= Tcl_FileReady (ptyPtr->file, TCL_EXCEPTION);
  }
  return present;
}

/*
 *----------------------------------------------------------------------
 *
 * ptyGetFile--
 *
 *	Called from Tcl_GetChannelFile to retrieve Tcl_Files from inside
 *	a file based channel.
 *
 * Results:
 *	The appropriate Tcl_File or NULL if not present. 
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */
static Tcl_File
ptyGetFile (instanceData, direction)
  ClientData instanceData;            /* The file state. */
  int direction;                      /* Which Tcl_File to retrieve? */
{
  ptyState *ptyPtr;

  ptyPtr = (ptyState *)instanceData;
  return ptyPtr->file;
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_OpenPtyChannel--
 *
 *   Opens a pty/tty pair and exec's a command on Unix systems.
 *
 * Results:
 *   The channel associated with the master side of the pty/tty pair.
 *
 * Side effects:
 *   May exec a command which could then do anything.
 *
 *----------------------------------------------------------------------
 */
Tcl_Channel
Tcl_OpenPtyChannel (interp, command, login, term, termp, winp)
  Tcl_Interp *interp;                   /* Interpreter for error reports.
                                           Can be NULL. */
  char *command;                        /* Command to exec. */
  int login;                            /* Set if we are creating a
                                           login shell. */
  char *term;                           /* Value to set TERM variable to. */
                                        /* Can be NULL. */
  struct termios *termp;                /* Initialised termios structure. */
                                        /* Can be NULL. */
  struct winsize *winp;                 /* Initialised winsize structure. */
                                        /* Can be NULL. */
{
  int pid[1], count, pgid, result;
  int master = -1;
  int slave = -1;
  char errSpace[128], channelName[16], argv0[256];
  Tcl_Channel channel;
  Tcl_DString buffer;
  Tcl_File file, errPipeIn, errPipeOut;
  ptyState *ptyPtr = (ptyState *)ckalloc ((unsigned)sizeof (ptyState));
  
/* XXX
 * munge command and arguments using Tcl_TranslateFileName as in
 * tclUnixPipe.c:TclSpawnPipeline
 */

  /*
   * Get a free and opened pty.
   */
  if (ptyGetPrivs () < 0) {
    if (interp != (Tcl_Interp *)NULL) {
      Tcl_AppendResult (interp, "couldn't restore set uid/gid privilages",
                        (char *)NULL);
      }
    goto error;
  }
  result = openpty (&master, &slave, ptyPtr->tline, termp, winp);
  if (ptyReleasePrivs () < 0) {
    if (interp != (Tcl_Interp *)NULL) {
      Tcl_AppendResult (interp, "couldn't release set uid/gid privilages",
                        (char *)NULL);
      }
    goto error;
  }
  if (result == -1) {
    if (interp != (Tcl_Interp *)NULL) {
      Tcl_AppendResult (interp, "couldn't open pty: no free pty's",
                        (char *) NULL);
    }
    goto error;
  }
  strcpy (ptyPtr->pline, ptyPtr->tline);
  ptyPtr->pline[5] = 'p';

  /*
   * Create a pipe that the child can use to return error
   * information if anything goes wrong.
   */
  if (TclCreatePipe (&errPipeIn, &errPipeOut) == 0) {
    if (interp != (Tcl_Interp *)NULL) {
      Tcl_AppendResult(interp, "couldn't create pipe: ",
                       Tcl_PosixError(interp), (char *) NULL);
    }
    goto error;
  }

  pid[0] = vfork();
  if (pid[0] == 0) {
    (void)close (master);
    if (login_tty (slave) < 0) {
      sprintf (errSpace, "%dforked process couldn't set up input/output: ",
               errno);
      TclWriteFile(errPipeOut, 1, errSpace, (int) strlen(errSpace));
      _exit(1);
    }
    for (count = getdtablesize(); count > 2; count--) {
      if (count != (int)Tcl_GetFileInfo (errPipeOut, NULL)) {
        (void)close (count);
      }
    }
    
    pgid = getpgrp();
    (void)ioctl (0, TIOCSPGRP, (char *)&pgid);
    (void)setpgid (0, pgid);
    (void)close (open (ptyPtr->tline, O_WRONLY, 0));
    (void)setpgid (0, pgid);

#if 0 /* Later */
    RestoreSignals();
#endif /* Later */

    if (!login) {
      execlp (command, NULL);
    } else {
      char *tmp;
      char *home = getenv ("HOME");

      if (home != (char *)NULL) {
        if (chdir (home) < 0) {
          sprintf(errSpace, "%dcouldn't chdir to \"%.150s\": ", errno, home);
          TclWriteFile(errPipeOut, 1, errSpace, (int) strlen(errSpace));
          _exit(1);
        }
      }
      if (term != (char *)NULL) {
        (void)setenv ("TERM", term, 1);
      } else {
        (void)setenv ("TERM", "dumb", 1);
      } 
      (void)strcpy (argv0, "-");
      tmp = strrchr (command, '/');
      if (tmp == NULL) {
        tmp = command;
      } else {
        tmp++;
      }
      (void)strncat (argv0, tmp, 255);
      execlp (command, argv0, NULL);
    }
    sprintf(errSpace, "%dcouldn't execute \"%.150s\": ", errno, command);
    TclWriteFile(errPipeOut, 1, errSpace, (int) strlen(errSpace));
    _exit(1);
  }
  if (pid[0] == -1) {
    if (interp != (Tcl_Interp *)NULL) {
      Tcl_AppendResult (interp, "couldn't fork child process: ",
                        Tcl_PosixError (interp), (char *)NULL);
    }
    goto error;
  } 

  /*
   * Read back error pipe.
   */
  TclCloseFile (errPipeOut);
  errPipeOut = NULL;

  count = TclReadFile (errPipeIn, 1, errSpace, (size_t)(sizeof (errSpace)-1));
  if (count > 0) {
    if (interp != (Tcl_Interp *)NULL) {
      char *end;
      errSpace[count] = 0;
      Tcl_SetErrno (strtol (errSpace, &end, 10));
      Tcl_AppendResult (interp, end, Tcl_PosixError (interp), (char *)NULL);
    }
    goto error;
  }
  TclCloseFile (errPipeIn);
  errPipeIn = NULL;

  /*
   * The command must have exec'd okay so fill in the Tcl stuff.
   */
  (void)close (slave);
  slave = -1;  

  Tcl_DetachPids (1, pid);

  sprintf (channelName, "pty%d", master);
  file = Tcl_GetFile ((ClientData)master, TCL_UNIX_FD);
  ptyPtr->file = file;
  channel = Tcl_CreateChannel (&ptyChannelType, channelName,
                               (ClientData) ptyPtr,
                               TCL_READABLE|TCL_WRITABLE);
  if (channel == (Tcl_Channel)NULL) {
    if (interp != (Tcl_Interp *)NULL) {
      Tcl_AppendResult (interp, "couldn't create channel \"", channelName,
                        "\": ", Tcl_PosixError (interp), (char *)NULL);
      Tcl_FreeFile (file);
      goto error;
    }
  }
      
  return channel;

error:
  if (errPipeIn) {
    TclCloseFile (errPipeIn);
  }
  if (errPipeOut) {
    TclCloseFile (errPipeOut);
  }
  (void)ptyGetPrivs ();
  if (master >= 0) {
    (void)close (master);
    (void)chmod (ptyPtr->pline, 0666);
    (void)chown (ptyPtr->pline, 0, 0);
  }
  if (slave >= 0) {
    (void)close (slave);
    (void)chmod (ptyPtr->tline, 0666);
    (void)chown (ptyPtr->tline, 0, 0);
  }
  (void)ptyReleasePrivs ();
  ckfree ((char *)ptyPtr);
  Tcl_ReapDetachedProcs();
  return NULL;
}

/*
 *----------------------------------------------------------------------
 *
 * ptyCmd proc
 *
 *   This procedure is used to process the "pty" Tcl command.
 *   See the user documentation on what is does.
 *
 * Results:
 *   A standard Tcl result.
 *
 * Side effects:
 *   See user documentation.
 *
 *----------------------------------------------------------------------
 */
int
ptyCmd (dummy, interp, argc, argv)
  ClientData dummy;
  Tcl_Interp *interp;
  int argc;
  char **argv;
{

  int i, login = 0;
  char *command, **argPtr;
  char *term = NULL;
  struct termios t;
  struct winsize w;
  Tcl_Channel channel;

  if (argc < 2) {
    wrong_num:
    Tcl_AppendResult (interp, "wrong # args: should be \"", argv[0],
                      " ?switches? command\"", (char *)NULL);
    return TCL_ERROR;
  }

  /*
   * Set up default termios structure
   */
  t.c_iflag = TTYDEF_IFLAG;
  t.c_oflag = TTYDEF_OFLAG;
  t.c_lflag = TTYDEF_LFLAG;
  t.c_cflag = TTYDEF_CFLAG;
  for (i = 0; i < NCCS; i++) {
    t.c_cc[i] = ttydefchars[i];
  }
  t.c_ispeed = TTYDEF_SPEED;
  t.c_ospeed = TTYDEF_SPEED;

  /*
   * Set up default winsize structure
   */
  w.ws_row = 0;
  w.ws_col = 0;
  w.ws_xpixel = 0;
  w.ws_ypixel = 0;
  
  argPtr = argv+1;
  argc--;
  while ((argc > 0) && ((argPtr[0][0] == '-') || (argPtr[0][0] == '+'))) {

    int got = 0;

    if ((argPtr[0][1] == 'l') &&
        (strcmp (argPtr[0], "-login") == 0)) {
      login++;
      argPtr++;
      argc--;
      continue;
    }

    if ((argPtr[0][1] == 't') &&
        (strcmp (argPtr[0], "-term") == 0)) {
      if (argc < 2) {
        Tcl_AppendResult (interp, "bad value for \"", argPtr[0],
                          "\": must be name",
                          (char *)NULL);
        return TCL_ERROR;
      }
      term = argPtr[1];
      argPtr++;
      argc--;
      argPtr++;
      argc--;
      continue;
    }

    /*
     * Termios Input Modes 
     */ 
    for (i = 0; i < IMODE_ELEMENTS; i++) {
      if ((argPtr[0][1] == imode_elements[i].sw[0]) &&
          (strcmp (&argPtr[0][1], imode_elements[i].sw) == 0)) {
        if (argPtr[0][0] == '+') {
          t.c_iflag |= imode_elements[i].value;
        } else {
          t.c_iflag &= !imode_elements[i].value;
        }
        got++;
        break;
      }
    }
    if (got) {
      argPtr++;
      argc--;
      continue;
    }

    /*
     * Termios Output Modes 
     */ 
    for (i = 0; i < OMODE_ELEMENTS; i++) {
      if ((argPtr[0][1] == omode_elements[i].sw[0]) &&
          (strcmp (&argPtr[0][1], omode_elements[i].sw) == 0)) {
        if (argPtr[0][0] == '+') {
          t.c_oflag |= omode_elements[i].value;
        } else {
          t.c_oflag &= !omode_elements[i].value;
        }
        got++;
        break;
      }
    }
    if (got) {
      argPtr++;
      argc--;
      continue;
    }

    /*
     * Termios Control Modes 
     */ 
    if ((argPtr[0][1] == 'c') &&
        (strncmp (argPtr[0], "-cs", 3) == 0) &&
        ((argPtr[0][3] >= '5') && (argPtr[0][3] <= '8'))) {
      t.c_cflag &= !CSIZE;
      if (argPtr[0][3] == '5') {
        t.c_cflag |= CS5;
      } else if (argPtr[0][3] == '6') {
        t.c_cflag |= CS6;
      } else if (argPtr[0][3] == '7') {
        t.c_cflag |= CS7;
      } else if (argPtr[0][3] == '8') {
        t.c_cflag |= CS8;
      }
      argPtr++;
      argc--;
      continue;
    }
    for (i = 0; i < CMODE_ELEMENTS; i++) {
      if ((argPtr[0][1] == cmode_elements[i].sw[0]) &&
          (strcmp (&argPtr[0][1], cmode_elements[i].sw) == 0)) {
        if (argPtr[0][0] == '+') {
          t.c_cflag |= cmode_elements[i].value;
        } else {
          t.c_cflag &= !cmode_elements[i].value;
        }
        got++;
        break;
      }
    }
    if (got) {
      argPtr++;
      argc--;
      continue;
    }

    /*
     * Termios Local Modes 
     */ 
    for (i = 0; i < LMODE_ELEMENTS; i++) {
      if ((argPtr[0][1] == lmode_elements[i].sw[0]) &&
          (strcmp (&argPtr[0][1], lmode_elements[i].sw) == 0)) {
        if (argPtr[0][0] == '+') {
          t.c_lflag |= lmode_elements[i].value;
        } else {
          t.c_lflag &= !lmode_elements[i].value;
        }
        got++;
        break;
      }
    }
    if (got) {
      argPtr++;
      argc--;
      continue;
    }
    
    /*
     * Termios Control Characters
     */
    for (i = 0; i < CCHAR_ELEMENTS; i++) {
      if ((argPtr[0][1] == cchar_elements[i].sw[0]) &&
          (strcmp (&argPtr[0][1], cchar_elements[i].sw) == 0) &&
          (argPtr[0][0] == '-')) {

        if ((argc >= 2) && ((int)strlen (argPtr[1]) == 1)) {
          t.c_cc[cchar_elements[i].value] = argPtr[1][0];
        } else if ((argc >= 2) && (strcmp (argPtr[1], "disable") == 0)) {
          t.c_cc[cchar_elements[i].value] = _POSIX_VDISABLE;
        } else {
          Tcl_AppendResult (interp, "bad value for \"", argPtr[0],
                            "\": must be character or disable",
                            (char *)NULL);
          return TCL_ERROR;
        }
        got++;
        break;
      }
    }
    if (got) {
      argPtr++;
      argc--;
      argPtr++;
      argc--;
      continue;
    }     
    if ((argPtr[0][1] == 'm') &&
        (strcmp (argPtr[0], "-min") == 0)) {

      if (argc >= 2) {
         int value;

        if (Tcl_GetInt (interp, argPtr[1], &value) == TCL_OK) {
          t.c_cc[VMIN] = value;
        } else if (strcmp (argPtr[1], "disable") == 0) {
            t.c_cc[VMIN] = _POSIX_VDISABLE;
        } else {
          return TCL_ERROR;
        }
      } else {
        Tcl_AppendResult (interp, "bad value for \"", argPtr[0],
                          "\": must be integer or disable",
                          (char *)NULL);
        return TCL_ERROR;
      }
      argPtr++;
      argc--;
      argPtr++;
      argc--;
      continue;
    }
    if ((argPtr[0][1] == 't') &&
        (strcmp (argPtr[0], "-time") == 0)) {

      if (argc >= 2) {
        int value;

        if (Tcl_GetInt (interp, argPtr[1], &value) == TCL_OK) {
          t.c_cc[VTIME] = value;
        } else if (strcmp (argPtr[1], "disable") == 0) {
            t.c_cc[VTIME] = _POSIX_VDISABLE;
        } else {
          return TCL_ERROR;
        }
      } else {
        Tcl_AppendResult (interp, "bad value for \"", argPtr[0],
                          "\": must be integer or disable",
                          (char *)NULL);
        return TCL_ERROR;
      }
      argPtr++;
      argc--;
      argPtr++;
      argc--;
      continue;
    }

    /*
     * Termios Speeds
     */
    if ((argPtr[0][1] == 'i') &&
        (strcmp (argPtr[0], "-ispeed") == 0)) {
      if (argc >= 2) {
        for (i = 0; i < BAUD_ELEMENTS; i++) {
          if ((argPtr[1][0] == baud_elements[i].sw[0]) &&
              (strcmp (argPtr[1], baud_elements[i].sw) == 0)) {
            t.c_ispeed = baud_elements[i].value;
            got++;
            break;
          } 
        }
      }
      if (got) {
        argPtr++;
        argc--;
        argPtr++;
        argc--;
        continue;
      } else {
        Tcl_AppendResult (interp, "bad value for \"", argPtr[0],
                          "\": must be one of ", (char *)NULL);
        for (i = 0; i < BAUD_ELEMENTS; i++) {
          Tcl_AppendResult (interp, baud_elements[i].sw, " ",
                            (char *)NULL);
        }
        return TCL_ERROR;
      }
    }
    if ((argPtr[0][1] == 'o') &&
        (strcmp (argPtr[0], "-ospeed") == 0)) {
      if (argc >= 2) {
        for (i = 0; i < BAUD_ELEMENTS; i++) {
          if ((argPtr[1][0] == baud_elements[i].sw[0]) &&
              (strcmp (argPtr[1], baud_elements[i].sw) == 0)) {
            t.c_ispeed = baud_elements[i].value;
            got++;
            break;
          } 
        }
      }
      if (got) {
        argPtr++;
        argc--;
        argPtr++;
        argc--;
        continue;
      } else {
        Tcl_AppendResult (interp, "bad value for \"", argPtr[0],
                          "\": must be one of ", (char *)NULL);
        for (i = 0; i < BAUD_ELEMENTS; i++) {
          Tcl_AppendResult (interp, baud_elements[i].sw, " ",
                            (char *)NULL);
        }
        return TCL_ERROR;
      }
    }

    /*
     * Winsize parameters
     */
    if ((argPtr[0][1] == 'r') &&
        (strcmp (argPtr[0], "-rows") == 0)) {
      if (argc >= 2) {
        int value;

        if (Tcl_GetInt (interp, argPtr[1], &value) == TCL_OK) {
          w.ws_row = value;
        } else {
          return TCL_ERROR;
        }
      } else {
        Tcl_AppendResult (interp, "bad value for \"", argPtr[0],
                          "\": must be integer",
                          (char *)NULL);
        return TCL_ERROR;
      }
      argPtr++;
      argc--;
      argPtr++;
      argc--;
      continue;
    }
    if ((argPtr[0][1] == 'c') &&
        (strcmp (argPtr[0], "-columns") == 0)) {
      if (argc >= 2) {
        int value;

        if (Tcl_GetInt (interp, argPtr[1], &value) == TCL_OK) {
          w.ws_col = value;
        } else {
          return TCL_ERROR;
        }
      } else {
        Tcl_AppendResult (interp, "bad value for \"", argPtr[0],
                          "\": must be integer",
                          (char *)NULL);
        return TCL_ERROR;
      }
      argPtr++;
      argc--;
      argPtr++;
      argc--;
      continue;
    }

    /*
     * Not a valid switch so must be an error
     */
    Tcl_AppendResult (interp, "bad switch \"", argPtr[0],
                      "\": must be -login, -term, ", (char *)NULL);
    for (i = 0; i < IMODE_ELEMENTS; i++) {
      Tcl_AppendResult (interp, "[+-]", imode_elements[i].sw,
                        ", ", (char *)NULL);
    }
    for (i = 0; i < OMODE_ELEMENTS; i++) {
      Tcl_AppendResult (interp, "[+-]", omode_elements[i].sw,
                        ", ", (char *)NULL);
    }
    Tcl_AppendResult (interp, "-cs5, -cs6, -cs7, -cs8, ", (char *)NULL);
    for (i = 0; i < CMODE_ELEMENTS; i++) {
      Tcl_AppendResult (interp, "[+-]", cmode_elements[i].sw,
                        ", ", (char *)NULL);
    }
    for (i = 0; i < LMODE_ELEMENTS; i++) {
      Tcl_AppendResult (interp, "[+-]", lmode_elements[i].sw,
                        ", ", (char *)NULL);
    }
    for (i = 0; i < CCHAR_ELEMENTS; i++) {
      Tcl_AppendResult (interp, "-", cchar_elements[i].sw,
                        " character, ", (char *)NULL);
    }
    Tcl_AppendResult (interp,
                      "-min time, -time time, ", 
                      "-ispeed baud rate, -ospeed baud rate, ",
                      "-rows number, -columns number",
                      (char *)NULL);
    return TCL_ERROR;
  }
      
  if (argc != 1) {
    goto wrong_num;
  }

  command = argPtr[0]; 

  channel = Tcl_OpenPtyChannel (interp, command, login, term, &t, &w);
  if (channel == (Tcl_Channel)NULL) {
    return TCL_ERROR;
  }
  Tcl_RegisterChannel (interp, channel);
  Tcl_AppendResult (interp, Tcl_GetChannelName (channel), (char *)NULL);
  return TCL_OK;

}
